### Produces:
### - Table 5 ("Moments of discretized Lévy processes for various period durations but constant maturity")
### - Table 6 ("Moments of discretized Brownian process for various maturities but constant period duration")

library(stats)
library(statmod)
library(parallel)

### Brownian and Poisson observationally equivalent at h=h0

sigma <- 0.022
mu <- 0.01
h0 <- 4

### 1/(4*sigma+mu/sigma)^2 = 3.40

num.cores <- detectCores()

simulate.moments <- function(h, mat, brownian, min.a = -0.8)
{
  gam <- 2
  lambda <- 0.005
  beta <- 0.8^h
  r <- 1.01^h-1

  delta <- h/mat*(1+r)-r
  
  if (brownian) {
    gu <- exp(sigma*sqrt(h))
    gd <- 1/gu
    pu <- 0.5 + mu/2/sigma*sqrt(h)
    pd <- 1-pu
  } else {
    p0 <- -log(0.5 + mu/2/sigma*sqrt(h0))/h0
    
    gd <- exp(-sigma*sqrt(h0))/(p0*h0/(1-exp(-p0*h0)))*(p0*h/(1-exp(-p0*h)))
    pu <- exp(-p0*h)
    pd <- 1-exp(-p0*h)

    ## Compute gu so that pu*gu+pd*gd the same than in the Brownian of same h
    x.brownian <- (0.5 + mu/2/sigma*sqrt(h))*exp(sigma*sqrt(h)) + (0.5 - mu/2/sigma*sqrt(h))*exp(-sigma*sqrt(h))
    gu <- (x.brownian - pd*gd)/pu
  }

  stopifnot(pu > 0 && pd > 0)

  stopifnot(beta*(pu*gu^(1-gam) + pd*gd^(1-gam))*(1+r) < 1)

  max.a <- 0
  ngrid.a <- 200
  convergence.tol <- 1e-5

  VG <- NULL
  VB <- NULL
  q1 <- NULL
  threshold <- NA

  ## Grids
  grid.a <- seq(min.a, max.a, length.out = ngrid.a)

  util <- function(c)
    ifelse(c <= 0, -1e8, c^(1-gam)/(1-gam))

  V <- function(a)
    max(VB, VG(a))

  q <- function(a1)
    {
      Erepay1 <- 0

      if (VG(a1/gu) >= VB)
        Erepay1 <- Erepay1 + pu

      if (VG(a1/gd) >= VB)
        Erepay1 <- Erepay1 + pd
      
      Erepay1/(1+r)*(delta+(1-delta)*q1(a1))
    }

  a1.dr <- function(a)
    {
      VG.objective <- function(a1)
        util(h+a*delta-(a1-(1-delta)*a)*q(a1)) + beta*(pu*gu^(1-gam)*V(a1/gu) + pd*gd^(1-gam)*V(a1/gd))
      
      zd <- max(min(threshold*gd, max.a), min.a)
      o1 <- optimize(VG.objective, lower = min.a, upper = zd, maximum = TRUE, tol = convergence.tol)
      o2 <- optimize(VG.objective, lower = zd, upper = max.a, maximum = TRUE, tol = convergence.tol)

      ifelse(o1$objective > o2$objective, o1$maximum, o2$maximum)
    }

  solve <- function(max.it = 10000)
    {
      ## Initial guess
      VB <<- util((1-lambda)*h)
      VG.points <- util(h + grid.a*r)
      q1.points <- rep(delta/(r+delta), ngrid.a)
      threshold <<- min(max(-lambda*h/r, min.a), max.a)

      init.spline <- function()
        {
          VG <<- splinefun(grid.a, VG.points, method = "monoH.FC")
          q1 <<- splinefun(grid.a, q1.points, method = "monoH.FC")
        }

      init.spline()

      ## Run iterations
      
      it <- 1
      err <- Inf

      VG.points.new <- numeric(ngrid.a)
      q1.points.new <- numeric(ngrid.a)

      while(err > convergence.tol && it <= max.it)
        {
          difference <- function(a)
            VG(a)-VB

          VB.new <- util((1-lambda)*h) + beta*(pu*gu^(1-gam) + pd*gd^(1-gam))*VB
          
          ## Compute VG
          VG.optimize <- function(a)
            {
              VG.objective <- function(a1)
                util(h+a*delta-(a1-(1-delta)*a)*q(a1)) + beta*(pu*gu^(1-gam)*V(a1/gu) + pd*gd^(1-gam)*V(a1/gd))
              
              zd <- max(min(threshold*gd, max.a), min.a)
              o1 <- optimize(VG.objective, lower = min.a, upper = zd, maximum = TRUE, tol = convergence.tol)
              o2 <- optimize(VG.objective, lower = zd, upper = max.a, maximum = TRUE, tol = convergence.tol)

              if (o1$objective > o2$objective)
                list(VG = o1$objective, q1 = q(o1$maximum))
              else
                list(VG = o2$objective, q1 = q(o2$maximum))
            }
          r <- mclapply(grid.a, VG.optimize, mc.cores = num.cores)
          VG.points.new <- sapply(r, function(x) x$VG)
          q1.points.new <- sapply(r, function(x) x$q1)

          ## Compute error
          err <- max(abs(VB.new-VB),
                     max(abs(VG.points.new-VG.points)))

          ## Copy
          VB <<- VB.new
          VG.points <- VG.points.new
          q1.points <- q1.points.new
          init.spline()

          ## Threshold
          if (difference(min.a) < 0 && difference(max.a) > 0)
            {
              threshold <<- uniroot(difference, lower = min.a, upper = max.a, tol = convergence.tol)$root
              cat(threshold, fill = TRUE)
            }
          else if (difference(min.a) >= 0)
            threshold <<- min.a
          else
            threshold <<- max.a

          ## Output
          cat("Iteration ", it, ": error=", err, fill = TRUE, sep = "")

          it <- it+1
        }

      if (threshold > a1.dr(threshold)/gd + convergence.tol)
        cat("Models with defaults", fill = TRUE)
      else
        cat("No defaults in this model", fill = TRUE)
      
      if (it == max.it+1)
        warning("Maximum iterations reached")

      threshold
    }

  simulation.has.default <- function(a0, length)
    {
      a <- a0
      
      for (i in 1:length)
        {
          if (VG(a) >= VB)
            {
              a <- a1.dr(a)/ifelse(runif(1) < pu, gu, gd)
            }
          else
            {
              return(TRUE)
            }
        }
      FALSE
    }

  default.prob <- function(nsims, sim.length = 40/h)
    {
      count <- 0
      for (i in 1:nsims)
        if (simulation.has.default(0, sim.length))
          count <- count + 1
      count / nsims
    }
  
  t <- solve()

  d <- mean(pvec(rep(NA, num.cores), function(x) default.prob(nsims = ceiling(5000/num.cores)), mc.cores = num.cores))
  
  list(threshold = t, default.prob = d, VG = VG)
}

### Produce Table 4

h <- c(h0, 2, 1)
min.a <- c(-0.8, -0.8, -0.7, -0.8, -0.7, -0.8)

results <- list()

for (i in 1:length(min.a))
  {
    this.h <- h[(i-1) %/% 2 + 1]
    this.brownian <- as.logical((i-1) %% 2)
    l <- simulate.moments(this.h, 4, this.brownian, min.a[i])
    l[["h"]] <- this.h
    l[["brownian"]] <- this.brownian
    results[[i]] <- l
  }

sink("vfi-2states-maturity.tex")
cat("\\begin{tabular}{l", paste(rep("|r", length(min.a)), collapse = ""), "}", fill = TRUE)
cat("\\hline \\hline", fill = TRUE)
cat("Period duration ($h$, in quarters) & ", paste(round(h, digit = 2), collapse = " & "), "\\\\ \\hline", fill = TRUE)
threshold.select <- function(i) formatC(abs(results[[i]]$threshold*100), format = "f", digits = 1)
default.prob.select <- function(i) formatC(results[[i]]$default.prob*100, format = "f", digits = 1)
cat("\\textit{Discretized Brownian process}", paste(rep("&", length(h))), "\\\\", fill = TRUE)
cat("Default threshold (debt-to-GDP, quarterly, \\%) & ", paste(sapply((1:length(h))*2, threshold.select), collapse = " & "), "\\\\", fill = TRUE)
cat("Default probability in 10 years (\\%) & ", paste(sapply((1:length(h))*2, default.prob.select), collapse = " & "), "\\\\ \\hline", fill = TRUE)
cat("\\textit{Discretized Poisson process}", paste(rep("&", length(h))), "\\\\", fill = TRUE)
cat("Default threshold (debt-to-GDP, quarterly, \\%) & ", paste(sapply((1:length(h))*2-1, threshold.select), collapse = " & "), "\\\\", fill = TRUE)
cat("Default probability in 10 years (\\%) & ", paste(sapply((1:length(h))*2-1, default.prob.select), collapse = " & "), "\\\\ \\hline \\hline", fill = TRUE)
cat("\\end{tabular}", fill = TRUE)
sink()

### Produce Table 5

mat <- c(1, 4, 20, 40)
min.a <- c(-0.8, -0.8, -1, -0.8)

results <- list()

for (i in 1:length(min.a))
  {
    l <- simulate.moments(1, mat[i], TRUE, min.a[i])
    l[["mat"]] <- mat[i]
    results[[i]] <- l
  }

sink("vfi-2states-maturity-constant-h.tex")
cat("\\begin{tabular}{l", paste(rep("|r", length(min.a)), collapse = ""), "}", fill = TRUE)
cat("\\hline \\hline", fill = TRUE)
cat("Debt maturity (Macaulay duration, in quarters) & ", paste(mat, collapse = " & "), "\\\\ \\hline", fill = TRUE)
threshold.select <- function(i) formatC(abs(results[[i]]$threshold*100), format = "f", digits = 1)
default.prob.select <- function(i) formatC(results[[i]]$default.prob*100, format = "f", digits = 1)
welfare.select <- function(i) formatC(results[[i]]$VG(0), format = "f", digits = 3)
cat("Default threshold (debt-to-GDP, quarterly, \\%) & ", paste(sapply(1:length(mat), threshold.select), collapse = " & "), "\\\\", fill = TRUE)
cat("Default probability in 10 years (\\%) & ", paste(sapply(1:length(mat), default.prob.select), collapse = " & "), "\\\\ \\hline", fill = TRUE)
cat("Welfare (starting from no debt) & ", paste(sapply(1:length(mat), welfare.select), collapse = " & "), "\\\\ \\hline \\hline", fill = TRUE)
cat("\\end{tabular}", fill = TRUE)
sink()
