## Produces Table 7 ("Simulation results for the Gaussian discrete approximation")
##
## Brownian process which is  discretized in time but not in value space
## (between 2 periods, growth follows a gaussian distribution)

library(stats)
library(statmod)
library(parallel)

sigma <- 0.022
mu <- 0.01-sigma^2/2 ## Correction for the Itô term

### 1/(4*sigma+mu/sigma)^2 = 3.40

num.cores <- detectCores()

simulate.moments <- function(h, min.a = -0.8)
{
  gam <- 2
  delta <- 0.005
  beta <- 0.8^h
  r <- 1.01^h-1

  max.a <- 0
  ngrid.a <- 25
  convergence.tol <- 1e-5

  VG <- NULL
  VB <- NULL

  threshold <- NA
  
  ## Grids
  grid.a <- seq(min.a, max.a, length.out = ngrid.a)

  util <- function(c)
    ifelse(c <= 0, -1e8, c^(1-gam)/(1-gam))

  V <- function(a)
    max(VB, VG(a))

  q <- function(a1)
    {
      if (a1 >= 0)
        return(1/(1+r))
      
      (1-pnorm(log(a1/threshold), mean = mu*h, sd = sigma*sqrt(h)))/(1+r)
    }

  a1.dr <- function(a, gauss.npoints = 16, truncation.window = 4)
    {
      gauss.points <- mu * h + gauss.quad(gauss.npoints)$nodes * truncation.window * sigma * sqrt(h)
      gauss.weights <- gauss.quad(gauss.npoints)$weights * truncation.window / sqrt(pi) / sqrt(2) * exp(-(truncation.window*gauss.quad(gauss.npoints)$nodes)^2/2)/(1-2*pnorm(-truncation.window*sigma*sqrt(h), sd = sigma*sqrt(h)))

      VG.objective <- function(a1)
          {
              EV1 <- 0
              for (i in 1:gauss.npoints)
                  EV1 <- EV1 + gauss.weights[i]*exp(gauss.points[i])^(1-gam)*V(a1/exp(gauss.points[i]))

              util(h+a-a1*q(a1)) + beta*EV1
          }
      
      optimize(VG.objective, lower = min.a, upper = max.a, maximum = TRUE, tol = convergence.tol)$maximum
    }

  solve <- function(max.it = 10000, gauss.npoints = 16, truncation.window = 4)
    {
      gauss.points <- mu * h + gauss.quad(gauss.npoints)$nodes * truncation.window * sigma * sqrt(h)
      gauss.weights <- gauss.quad(gauss.npoints)$weights * truncation.window / sqrt(pi) / sqrt(2) * exp(-(truncation.window*gauss.quad(gauss.npoints)$nodes)^2/2)/(1-2*pnorm(-truncation.window*sigma*sqrt(h), sd = sigma*sqrt(h)))

      ## Initial guess
      VB <<- util((1-delta)*h)
      VG.points <- util(h + grid.a*r)

      init.spline <- function()
        VG <<- splinefun(grid.a, VG.points, method = "monoH.FC")

      init.spline()

      threshold <<- -delta*h/r

      ## Run iterations
      
      it <- 1
      err <- Inf

      VG.points.new <- numeric(ngrid.a)

      while(err > convergence.tol && it <= max.it)
        {
          difference <- function(a)
            VG(a)-VB

          EVB1 <- 0
          for (i in 1:gauss.npoints)
              EVB1 <- EVB1 + gauss.weights[i]*exp(gauss.points[i])^(1-gam)*VB
          VB.new <- util((1-delta)*h) + beta*EVB1
          
          ## Compute VG
          VG.optimize <- function(a)
            {
              VG.objective <- function(a1)
                  {
                      EV1 <- 0
                      for (i in 1:gauss.npoints)
                          EV1 <- EV1 + gauss.weights[i]*exp(gauss.points[i])^(1-gam)*V(a1/exp(gauss.points[i]))

                      util(h+a-a1*q(a1)) + beta*EV1
                  }
              
              optimize(VG.objective, lower = min.a, upper = max.a, maximum = TRUE, tol = convergence.tol)$objective
            }
          VG.points.new <- as.numeric(mclapply(grid.a, VG.optimize, mc.cores = num.cores))

          ## Compute error
          err <- max(abs(VB.new-VB),
                     max(abs(VG.points.new-VG.points)))

          ## Copy
          VB <<- VB.new
          VG.points <- VG.points.new
          init.spline()

          ## Threshold
          threshold <<- uniroot(difference, lower = min.a, upper = max.a, tol = convergence.tol)$root
          cat(threshold, fill = TRUE)

          ## Output
          cat("Iteration ", it, ": error=", err, fill = TRUE, sep = "")

          it <- it+1
        }
      
      if (it == max.it+1)
        warning("Maximum iterations reached")

      threshold
    }

  simulation.has.default <- function(a0, length)
    {
      a <- a0
      
      for (i in 1:length)
        {
          if (VG(a) >= VB)
            {
              a <- a1.dr(a)/exp(rnorm(1, mean = mu*h, sd = sigma*sqrt(h)))
            }
          else
            {
              return(TRUE)
            }
        }
      FALSE
    }

  default.prob <- function(nsims = 1000, sim.length = 40/h)
    {
      count <- 0
      for (i in 1:nsims)
        if (simulation.has.default(0, sim.length))
          count <- count + 1
      count / nsims
    }
  
  t <- solve()

  d <- mean(pvec(rep(NA, num.cores), function(x) default.prob(nsims = ceiling(5000/num.cores)), mc.cores = num.cores))
  
  list(threshold = t, default.prob = d)
}

h <- c(4, 2, 1, 0.33, 0.08, 0.01)
min.a <- c(-0.6, -0.6, -0.6, -0.6, -0.6, -0.8)

results <- list()

for (i in 1:length(min.a))
  {
    this.h <- h[i]
    l <- simulate.moments(this.h, min.a[i])
    l[["h"]] <- this.h
    results[[i]] <- l
  }

## sink("vfi-2states.tex")
## cat("\\begin{tabular}{l", paste(rep("|r", length(min.a)), collapse = ""), "}", fill = TRUE)
## cat("\\hline \\hline", fill = TRUE)
## cat("Period duration ($h$, in quarters) & ", paste(round(h, digit = 2), collapse = " & "), "\\\\ \\hline", fill = TRUE)
## threshold.select <- function(i) formatC(abs(results[[i]]$threshold*100), format = "f", digits = 1)
## default.prob.select <- function(i) formatC(results[[i]]$default.prob*100, format = "f", digits = 1)
## cat("\\textit{Discretized Brownian process}", paste(rep("&", length(h))), "\\\\", fill = TRUE)
## cat("Default threshold (debt-to-GDP, quarterly, \\%) & ", paste(sapply((1:length(h))*2, threshold.select), collapse = " & "), "\\\\", fill = TRUE)
## cat("Default probability in 10 years (\\%) & ", paste(sapply((1:length(h))*2, default.prob.select), collapse = " & "), "\\\\ \\hline", fill = TRUE)
## cat("\\textit{Discretized Poisson process}", paste(rep("&", length(h))), "\\\\", fill = TRUE)
## cat("Default threshold (debt-to-GDP, quarterly, \\%) & ", paste(sapply((1:length(h))*2-1, threshold.select), collapse = " & "), "\\\\", fill = TRUE)
## cat("Default probability in 10 years (\\%) & ", paste(sapply((1:length(h))*2-1, default.prob.select), collapse = " & "), "\\\\ \\hline \\hline", fill = TRUE)
## cat("\\end{tabular}", fill = TRUE)
## sink()
